<xsl:transform
 xmlns:xsl="http://www.w3.org/1999/XSL/Transform"
 xmlns:txsl="http://www.w3.org/1999/XSL/Transform/target"
 version="1.0">

  <xsl:strip-space elements="*"/>
  <xsl:output indent="yes"/>
  <xsl:namespace-alias stylesheet-prefix="txsl" result-prefix="xsl"/>

  <xsl:template match="test">
    <txsl:template name="{@name}">
      <xsl:call-template name="process-block"/>
    </txsl:template>
  </xsl:template>

  <xsl:template match="return">
    <result>
      <xsl:value-of select="@code"/>
    </result>
  </xsl:template>

  <xsl:template match="*[descendant::return]" mode="metastatement" priority="1">
    <txsl:variable name="result">
      <xsl:apply-templates select="."/>
    </txsl:variable>
    <txsl:choose>
      <txsl:when test="$result/result">
        <txsl:copy-of select="$result/result"/>
      </txsl:when>
      <txsl:otherwise>
        <xsl:variable name="skip" select="count(preceding-sibling::*) + 1"/>
        <xsl:for-each select="..">
          <xsl:call-template name="process-block">
            <xsl:with-param name="skip" select="$skip"/>
          </xsl:call-template>
        </xsl:for-each>
      </txsl:otherwise>
    </txsl:choose>
  </xsl:template>
  
  <xsl:template match="while" mode="metastatement" priority="10">
    <txsl:if test="{@test}">
      <txsl:call-template name="{generate-id(.)}">
        <xsl:for-each select="ancestor-or-self::*/preceding-sibling::*[self::xsl:variable|self::xsl:param]">
          <txsl:with-param name="{@name}" select="${@name}"/>
        </xsl:for-each>
      </txsl:call-template>
    </txsl:if>
    <txsl:template name="{generate-id(.)}">
      <xsl:for-each select="ancestor-or-self::*/preceding-sibling::*[self::xsl:variable|self::xsl:param]">
        <txsl:param name="{@name}"/>
      </xsl:for-each>
      <xsl:call-template name="process-block"/>
      <xsl:variable name="next-statements" select="next"/>
      <txsl:choose>
        <xsl:if test="descendant::return">
          <txsl:when test="$result/result"/>
        </xsl:if>
        <txsl:when test="{@test}">
          <txsl:call-template name="{generate-id(.)}">
            <xsl:for-each select="ancestor-or-self::*/preceding-sibling::*[self::xsl:variable|self::xsl:param]">
              <xsl:variable name="varname" select="string(@name)"/>
              <xsl:if test="not($next-statements[@name=$varname])">
                <txsl:with-param name="{@name}" select="${@name}"/>
              </xsl:if>
            </xsl:for-each>
            <xsl:for-each select="next">
              <txsl:with-param name="{@name}" select="{@select}"/>
            </xsl:for-each>
          </txsl:call-template>
        </txsl:when>
        <txsl:otherwise>
          <txsl:call-template name="after_{generate-id(.)}">
            <xsl:for-each select="ancestor-or-self::*/preceding-sibling::*[self::xsl:variable|self::xsl:param]">
              <xsl:variable name="varname" select="string(@name)"/>
              <xsl:if test="not($next-statements[@name=$varname])">
                <txsl:with-param name="{@name}" select="${@name}"/>
              </xsl:if>
            </xsl:for-each>
            <xsl:for-each select="next">
              <txsl:with-param name="{@name}" select="{@select}"/>
            </xsl:for-each>
          </txsl:call-template>
        </txsl:otherwise>
      </txsl:choose>
    </txsl:template>
    <txsl:template name="after_{generate-id(.)}">
      <xsl:for-each select="ancestor-or-self::*/preceding-sibling::*[self::xsl:variable|self::xsl:param]">
        <txsl:param name="{@name}"/>
      </xsl:for-each>
      <xsl:variable name="skip" select="count(preceding-sibling::*) + 1"/>
      <xsl:for-each select="..">
        <xsl:call-template name="process-block">
          <xsl:with-param name="skip" select="$skip"/>
        </xsl:call-template>
      </xsl:for-each>
    </txsl:template>
  </xsl:template>

  <xsl:template match="next"/>
  
  <xsl:template match="text()">
    <xsl:copy/>
  </xsl:template>
  
  <xsl:template match="*">
    <xsl:copy>
      <xsl:copy-of select="text()|@*"/>
      <xsl:call-template name="process-block"/>
    </xsl:copy>
  </xsl:template>
  
  <xsl:template name="process-block">
    <xsl:param name="skip">0</xsl:param>

    <!-- Find first metastatement. -->
    <xsl:variable name="metastatement-location">
      <xsl:variable name="locations">
        <xsl:for-each select="*">
          <xsl:if test="position() &gt; $skip">
            <xsl:if test="self::while|descendant::return">
              <loc>
                <xsl:value-of select="position()"/>
              </loc>
            </xsl:if>
          </xsl:if>
        </xsl:for-each>
      </xsl:variable>
      <xsl:choose>
        <xsl:when test="count($locations/loc) &gt; 0">
          <xsl:value-of select="$locations/loc[1]"/>
        </xsl:when>
        <xsl:otherwise>0</xsl:otherwise>
      </xsl:choose>
    </xsl:variable>

    <xsl:choose>
      <xsl:when test="$metastatement-location &gt; 0">
        <!-- Process instructions preceding the metastatement -->
        <xsl:apply-templates select="*[position() &gt; $skip and position() &lt; $metastatement-location]"/>
        <!-- Process the metastatement.  The metastatement will process its following instructions itself -->
        <xsl:apply-templates select="*[position() = $metastatement-location]" mode="metastatement"/>
      </xsl:when>
      <xsl:otherwise>
        <xsl:apply-templates select="*[position() &gt; $skip]"/>
      </xsl:otherwise>
    </xsl:choose>
  </xsl:template>

  <xsl:template name="copy-template">
    <xsl:copy>
      <xsl:copy-of select="text()|@*"/>
      <xsl:for-each select="*">
         <xsl:call-template name="skip-templates"/>
      </xsl:for-each>
    </xsl:copy>
  </xsl:template>

  <xsl:template name="skip-templates">
    <xsl:if test="not(self::xsl:template)">
      <xsl:copy>
        <xsl:copy-of select="text()|@*"/>
        <xsl:for-each select="*">
           <xsl:call-template name="skip-templates"/>
        </xsl:for-each>
      </xsl:copy>
    </xsl:if>
  </xsl:template>

  <xsl:template match="/">
    <xsl:variable name="results">
      <xsl:apply-templates select="*"/>
    </xsl:variable>
    
    <txsl:transform version="1.0">
      <xsl:for-each select="$results//xsl:template">
        <xsl:call-template name="copy-template"/>
      </xsl:for-each>
    </txsl:transform>
  </xsl:template>

</xsl:transform>
